function [h] = findhypothesis_seq (X, Y, u, theta, max_col, htype);

% Find a high gain sequence boosting stump
%
% Input
%    X: (1,n) cellarray of n sequences.
%    Y: (n,1) sample labels (1 or -1).
%    u: (n,1) sample weights (>= 0).
%    theta: Minimum required gain.
%    max_col: Maximum number of constraint violating columns/stumps to produce.
%    htype: 1: 1-class and 1.5-class, outputs {0, 1}
%           2: 2-class, outputs {-1, 1}
%
% Output
%    h: (1,p), p <= max_col, cellarray of structures with elements
%       .h: classifier function.
%       .hi: additional information.
%       .GY: classifier response, one (n,1) column.

% The maximum number of subgraph patterns to return.
boostN = 1;
if max_col > 0
	boostN = max_col;
end

if theta < 0.0
	theta = 0.0;
end

% subg: (1,p) cellarray of sequences
% ybase: (1,p) basic response value (normally constant)
% GY: (n,p) response on the n training patterns
[subg, ybase, GY] = mexprefixspan (X, 2, [1 0], Y, u, theta, boostN, htype);
disp(['PrefixSpan mining returned ', num2str(length(subg)), ' subgraphs']);

if length(subg) == 0
	h={};
else
	h={};
	for i=1:length(subg)
		h{i}=[];
		h{i}.h = @(S) seq_stump_classifier (S, subg{i}, ybase(i), htype);
		h{i}.hi = subg{i};
		h{i}.ybase = ybase(i);
		h{i}.GY = GY(:,i);	% subsequence response on training data
	end
end


% ybase: 1 for 1.5-class LPBoosting, -1/1 for 2-class LPBoost
function [y] = seq_stump_classifier (S, subg, ybase, htype);

y=zeros(length(S),1);
for i=1:length(S)
	cury = seq_test (subg, S{i});

	if htype == 1
		y(i) = cury * ybase;
	elseif htype == 2
		y(i) = (2*cury - 1) * ybase;
	end
end


function [y] = seq_test (sub, S);

j=1;
for i=1:length(sub)
	while length(intersect(sub{i},S{j})) ~= length(sub{i})
		% No match, advance larger sequence element pointer
		j=j+1;

		% End of larger sequence?
		if j==(length(S)+1)
			y = 0;
			return;
		end
	end

	% Reached end of subsequence?
	j = j + 1;
	if i==length(sub)
		% Success
		y = 1;
		return;
	end

	% Case: partial match, but end of S
	if j==(length(S)+1)
		y = 0;
		return;
	end
end

